prefix=/usr/local
exec_prefix=${prefix}

PATH=${exec_prefix}/sbin:$PATH
export PATH

LD_LIBRARY_PATH=${exec_prefix}/lib:$LD_LIBRARY_PATH
export LD_LIBRARY_PATH

GLUSTERD_WORKDIR=/var/lib/glusterd
export GLUSTERD_WORKDIR

PYTHONPATH=/usr/lib/python2.6/site-packages:$PYTHON_PATH
export PYTHONPATH

PYTHON=/usr/bin/python
export PYTHON

PYTHONPATH=/usr/lib/python2.6/site-packages:$PYTHON_PATH
export PYTHONPATH
