int notifier_with_return_list_init () { return 0; }
int notifier_with_return_list_notify () { return 0; }
int notifier_with_return_list_add () { return 0; }
int notifier_list_init () { return 0; }
int notifier_list_notify () { return 0; }
int notifier_list_add () { return 0; }
int monitor_protocol_event () { return 0; }
int block_job_cancel_sync () { return 0; }
int block_job_iostatus_reset () { return 0; }
int vm_stop () { return 0; }
int qemu_get_aio_context () { return 0; }

